--- @class world
local world = {}

local sky = require "sky"
local tiles = require "tiles"
local serializer = require "serializer"
local object = require "object"

function world:getchunkcoords(tilex, tiley)
    return math.floor(tilex/50), math.floor(tiley/50)
end

function world:getchunk(x, y)
    local chunkid
    if type(x) == "string" then
        chunkid = x
    else
        chunkid = x.."_"..y
    end
    return self.chunks[chunkid], chunkid
end

function world:unloadchunk(x, y)
    local chunk, chunkid = self:getchunk(x, y)
    self:savechunk(x, y, function(success)
        if success then
            -- print "unloaded"
            self.chunks[chunkid] = nil
        end
    end)
end

function world:savechunksync(x, y)
    local chunk, chunkid = self:getchunk(x, y)
    local chunktemps = getmetatable(chunk).__index
    if not self.savefile then
        chunktemps.saving = true
        return false
    end
    serializer.save(chunk, self.savefile.."/chunkdata/"..chunkid..".table")
    return true
end

function world:savechunk(x, y, callback)
    local chunk, chunkid = self:getchunk(x, y)
    local chunktemps = getmetatable(chunk).__index
    chunktemps.saving = true
    if not self.savefile then
        return callback(false, "nosave")
    end
    asyncfs.write(self.savefile.."/chunkdata/"..chunkid..".table", serializer.serialize(chunk), function(event)
        if not event.success then
            print(event.errormessage)
        end
        callback(event.success, event.errormessage)
    end)
    return
end

function world:loadchunklazy(x, y, dontsimulate, callback)
    local generator = require(self.generator)
    local chunkid = x.."_"..y
    local chunk = self.chunks[chunkid]
    local chunktemps
    local shouldgenerate = false
    local fixobjects = false
    local function onload()
        if not chunk then
            fixobjects = false
            chunk = {
                tiles = {},
                objects = {},
            }
            shouldgenerate = true
        end
        if not chunktemps then
            chunktemps = getmetatable(chunk).__index
        end
        self.chunks[chunkid] = setmetatable(chunk, {__index = chunktemps})
        if shouldgenerate then
            generator(self, {
                x = x * 50,
                y = y * 50,
                width = 50,
                height = 50
            }, self.generatorsettings)
            -- print(x, y)
        end
        if dontsimulate == nil then
            chunktemps.simulate = chunktemps.simulate or false
        elseif dontsimulate == true then
            chunktemps.simulate = false
        elseif dontsimulate == false then
            chunktemps.simulate = true
        end
        chunktemps.saving = false
        if fixobjects then
            for key, value in pairs(chunk.objects) do
                chunk.objects[key] = object(value)
            end
        end
    end
    if chunk then
        onload()
    else
        chunktemps = {
            x = x,
            y = y,
            id = chunkid,
            simulate = false,
            needstosave = false
        }
        if self.savefile then
            asyncfs.read(self.savefile.."/chunkdata/"..chunkid..".table", function(event)
                if event.contents then
                    chunk = serializer.deserialize(event.contents)
                    fixobjects = true
                    onload()
                else
                    onload()
                end
            end)
        else
            onload()
        end
    end
    callback(self.chunks[chunkid], chunkid)
end

function world:loadchunksync(x, y, dontsimulate)
    local generator = require(self.generator)
    local chunkid = x.."_"..y
    local chunk = self.chunks[chunkid]
    local chunktemps
    local shouldgenerate = false
    local fixobjects = false
    if not chunk then
        chunktemps = {
            x = x,
            y = y,
            id = chunkid,
            simulate = false,
            needstosave = false
        }
        if self.savefile then
            chunk = serializer.load(self.savefile.."/chunkdata/"..chunkid..".table")
            fixobjects = true
        end
        if not chunk then
            fixobjects = false
            chunk = {
                tiles = {},
                objects = {},
            }
            shouldgenerate = true
        end
        self.chunks[chunkid] = setmetatable(chunk, {__index = chunktemps})
    end
    if not chunktemps then
        chunktemps = getmetatable(chunk).__index
    end
    if shouldgenerate then
        generator(self, {
            x = x * 50,
            y = y * 50,
            width = 50,
            height = 50
        }, self.generatorsettings)
        -- print(x, y)
    end
    if dontsimulate == nil then
        chunktemps.simulate = chunktemps.simulate or false
    elseif dontsimulate == true then
        chunktemps.simulate = false
    elseif dontsimulate == false then
        chunktemps.simulate = true
    end
    chunktemps.saving = false
    if fixobjects then
        for key, value in pairs(chunk.objects) do
            chunk.objects[key] = object(value)
        end
    end
    return self.chunks[chunkid], chunkid
end

function world:set(x, y, value)
    local chunkx, chunky = self:getchunkcoords(x, y)
    local chunk = self:loadchunksync(chunkx, chunky)
    chunk.tiles[math.floor(x).."_"..math.floor(y)] = value
end

function world:get(x, y, loadchunk)
    if loadchunk == nil then
        loadchunk = true
    end
    local chunkx, chunky = self:getchunkcoords(x, y)
    local chunk
    if loadchunk then
        chunk = self:loadchunksync(chunkx, chunky)
    else
        chunk = self:getchunk(chunkx, chunky)
        if not chunk then
            return tiles.air
        end
    end
    local id = chunk.tiles[math.floor(x).."_"..math.floor(y)]
    if id then
        return tiles[id]
    end
    return tiles.air
end

function world:autoloadchunks(x, y)
    -- print ("-", x, y)
    for chunkid, chunk in pairs(self.chunks) do
        if math.abs(chunk.x - x) > 2 or math.abs(chunk.y - y) > 2 then
            self:unloadchunk(chunkid)
            -- print("unloaded", chunkid)
        end
    end
    for xpos = -2, 2 do
        for ypos = -2, 2 do
            self:loadchunklazy(xpos+x, ypos+y, math.abs(xpos) == 2 or math.abs(ypos) == 2, function()
            
            end)
            -- print("loaded", xpos+x, ypos+y, xpos, ypos, math.abs(xpos) == 2 or math.abs(ypos) == 2, xpos == 0 and ypos == 0)
        end
    end
end

function world:getid(x, y)
    local tile = self:get(x, y)
    if tile then
        return tile.id
    end
    return nil
end



function world:getspace(x, y)
    return math.floor(x/8), math.floor(y/8)
end

function world:x(x)
    if not self.metax[x] then
        self.metax[x] = {}
    end
    return self.metax[x]
end

function world:y(y)
    if not self.metay[y] then
        self.metay[y] = {}
    end
    return self.metay[y]
end

function world:instantspawn(newobj)
    local chunkx, chunky = self:getchunkcoords(self:getspace(newobj.x, newobj.y))
    local chunk, chunkid = self:loadchunksync(chunkx, chunky)
    chunk.objects[newobj.uid] = newobj
    self.objects[newobj.uid] = chunkid
    return newobj
end

function world:instantdestroy(id)
    local chunkid = self.objects[id]
    if not chunkid then
        return
    end
    local chunk = self.chunks[chunkid]
    chunk.objects[id] = nil
end

function world:spawn(x, y, objecttype)
    self.uid = self.uid + 1
    local newobj = object {
        x = x,
        y = y,
        type = objecttype,
        uid = self.uid
    }
    self:queue(function()
        local o = self:instantspawn(newobj)
        if o.init then
            o:init(self)
        end
    end)
    return newobj
end

function world:destroy(id_or_object)
    local id = id_or_object
    if type(id_or_object) == "table" then
        id = id_or_object.uid
    end
    self:queue(function()
        self:instantdestroy(id)
    end)
end

local debuglist = {}

function world.pushdebug(key, prefix)
    if type(key) == "function" then
        local prefix = prefix or ""
        debuglist[#debuglist+1] = {key, prefix}
    else
        local prefix = prefix or ""
        debuglist[#debuglist+1] = {key, prefix..key}
    end
end

world.pushdebug("seed")
world.pushdebug("camerax")
world.pushdebug("cameray")
world.pushdebug("time")
world.pushdebug("uid")
world.pushdebug(function(self)
    local i = 0
    local j = 0
    for key, value in pairs(self.chunks) do
        i = i + 1
        if value.simulate then
            j = j + 1
        end
    end
    return i..":"..j
end, "loadedchunks")
world.pushdebug("mousex")
world.pushdebug("mousey")

function world:debugstring()
    local str = {}
    for index, value in ipairs(debuglist) do
        if type(value[1]) == "function" then
            str[#str+1] = value[2]..": "..tostring(value[1](self))
        else
            str[#str+1] = value[2]..": "..tostring(self[value[1]])
        end
    end
    return table.concat(str, "\n")
end

function world:viewWidth()
    return love.graphics.getWidth() / self.scale
end

function world:viewHeight()
    return love.graphics.getHeight() / self.scale
end

function world:getobj(id)
    local chunkid = self.objects[id]
    if not chunkid then
        return
    end
    local chunk = self.chunks[chunkid]
    local obj = chunk.objects[id]
    return obj
end

function world:queue(func)
    self.eventqueue[#self.eventqueue+1] = func
end

function world:updateobject(dt, events, obj)
    if obj.update then
        obj:update(self, dt)
    end
    if self.allowinput then
        for i = 1, #events do
            local event = events[i]
            if obj[event.type] then
                obj[event.type](obj, self, unpack(event.args))
            end
        end
    end
end

function world:mouseIsDown(button)
    return love.mouse.isDown(button) and self.allowinput
end

function world:keyIsDown(key)
    return love.keyboard.isDown(key) and self.allowinput
end

function world:isDown(control)
    if type(control) == "number" then
        return self:mouseIsDown(control)
    else
        return self:keyIsDown(control)
    end
end

function world:update(dt, events, allowinput)
    self.allowinput = allowinput
    self.controls.left = self:isDown "left" or self:isDown "a"
    self.controls.right = self:isDown "right" or self:isDown "d"
    self.controls.up = self:isDown "up" or self:isDown "w" or self:isDown "space"
    self.controls.down = self:isDown "down" or self:isDown "s"
    self.controls.attack = self:isDown(0)
    self.controls.interact = self:isDown(1)
    self.controls.pick = self:isDown(2)
    if self.allowinput then
        local tx = math.floor(-self.camerax * self.scale) / self.scale
        local ty = math.floor(-self.cameray * self.scale) / self.scale
        self.mousex = love.mouse.getX() / self.scale - tx
        self.mousey = love.mouse.getY() / self.scale - ty
    end
    self.time = self.time + 1
    for id, chunk in pairs(self.chunks) do
        if chunk.simulate then
            for uid, object in pairs(chunk.objects) do
                self:updateobject(dt, events, object)
            end
        end
    end
    local prevchunkx, prevchunky = self:getchunkcoords(self:getspace(self.player.x, self.player.y))
    self:updateobject(dt, events, self.player)
    local newchunkx, newchunky = self:getchunkcoords(self:getspace(self.player.x, self.player.y))
    if prevchunkx ~= newchunkx or prevchunky ~= newchunky then
        self:autoloadchunks(newchunkx, newchunky)
    end
    for id, chunk in pairs(self.chunks) do
        for uid, object in pairs(chunk.objects) do
            local newchunk, chunkid = self:getchunk(self:getchunkcoords(self:getspace(object.x, object.y)))
            if chunkid ~= id then
                -- print(uid, "moved from", id, "to", chunkid)
                chunk.objects[uid] = nil
                newchunk.objects[uid] = object
                self.objects[uid] = chunkid
            end
        end
    end
    for index, value in ipairs(self.eventqueue) do
        value()
        self.eventqueue[index] = nil
    end
end

local tileset = love.graphics.newImage("sprites/tiles.png")

function world:draw()
    love.graphics.push()
    love.graphics.scale(self.scale)
    sky.draw(self, self.camerax/8, self.cameray/8)
    local tx = math.floor(-self.camerax * self.scale) / self.scale
    local ty = math.floor(-self.cameray * self.scale) / self.scale
    love.graphics.translate(tx, ty)
    for id, chunk in pairs(self.chunks) do
        for uid, object in pairs(chunk.objects) do
            if object.draw then
                object:draw(self)
                love.graphics.setColor(1, 1, 1)
            end
        end
    end
    self.player:draw(self)
    local tcamx = math.floor(self.camerax/8)
    local tcamy = math.floor(self.cameray/8)
    for x = tcamx, tcamx + love.graphics.getWidth()/8/self.scale + 1 do
        for y = tcamy, tcamy + love.graphics.getHeight()/8/self.scale + 1 do
            local tile = self:get(x, y, false)
            if tile then
                local quad = love.graphics.newQuad(tile.sx, tile.sy, tile.sw, tile.sh, tileset:getWidth(), tileset:getHeight())
                love.graphics.draw(tileset, quad, x*8, y*8)
            end
        end
    end
    for id, chunk in pairs(self.chunks) do
        for uid, object in pairs(chunk.objects) do
            if object.drawtop then
                object:drawtop(self)
                love.graphics.setColor(1, 1, 1)
            end
        end
    end
    self.player:drawtop(self)
    if self.debug then
        love.graphics.setColor(1, 0, 0)
        love.graphics.setLineWidth(2)
        love.graphics.rectangle("line", math.floor(self.camerax/8/50+1)*8*50, math.floor(self.cameray/8/50)*8*50, 8*50, 8*50)
        love.graphics.rectangle("line", math.floor(self.camerax/8/50+1)*8*50, math.floor(self.cameray/8/50+1)*8*50, 8*50, 8*50)
        love.graphics.rectangle("line", math.floor(self.camerax/8/50)*8*50, math.floor(self.cameray/8/50+1)*8*50, 8*50, 8*50)
        love.graphics.rectangle("line", math.floor(self.camerax/8/50)*8*50, math.floor(self.cameray/8/50)*8*50, 8*50, 8*50)
        love.graphics.rectangle("fill", self.mousex-1, self.mousey-1, 2, 2)
        love.graphics.rectangle("line", math.floor(self.mousex/8)*8, math.floor(self.mousey/8)*8, 8, 8)
    end
    love.graphics.translate(-tx, -ty)
    if self.debug then
        local chunkx, chunky = self:getchunkcoords(self:getspace(self.player.x, self.player.y))
        for x = 0, 8 do
            for y = 0, 8 do
                local chunk, chunkid = self:getchunk(chunkx-4+x, chunky-4+y)
                if chunk then
                    local b = 0
                    if chunk.x == chunkx and chunk.y == chunky then
                        b = 1
                    end
                    if chunk.simulate then
                        love.graphics.setColor(1, b, 0)
                    else
                        if chunk.saving then
                            love.graphics.setColor(0, b, 1)
                        else
                            love.graphics.setColor(0, 1, b)
                        end
                    end
                else
                    love.graphics.setColor(0.5, 0.5, 0.5)
                end
                love.graphics.rectangle("fill", x*4, y*4 - 36 + self:viewHeight(), 4, 4)
            end
        end
    end
    love.graphics.setColor(1, 1, 1)
    for id, chunk in pairs(self.chunks) do
        for uid, object in pairs(chunk.objects) do
            if object.drawui then
                object:drawui(self)
                love.graphics.setColor(1, 1, 1)
            end
        end
    end
    self.player:drawui(self)
    love.graphics.setColor(1, 1, 1)
    love.graphics.pop()
    if self.debug then
        love.graphics.print(self:debugstring())
    end
end

function world:init(savefile)
    self.seed = love.math.random() * 1000
    self.player = object({
        x = 0,
        y = 0
    }, true)
    self.controls = {}
    self.time = 0
    self.metax = {}
    self.metay = {}
    self.objects = {}
    self.uid = 0
    self.camerax = 0
    self.cameray = 0
    self.eventqueue = {}
    self.chunks = {}
    self.mousex = 0
    self.mousey = 0
    self.generator = "generators.default"
    self.generatorsettings = {}
    self.savefile = savefile
    self.debug = false
    self.allowinput = false
    self.player:init(self)
    local chunkx, chunky = self:getchunkcoords(self:getspace(self.player.x, self.player.y))
    self:autoloadchunks(chunkx, chunky)
end

function world:load(savefile)
    self.savefile = savefile
    self.chunks = {}
    self.objects = serializer.load(savefile.."/objects.table")
    self.metax = serializer.load(savefile.."/metax.table")
    self.metay = serializer.load(savefile.."/metay.table")
    self.player = object(serializer.load(savefile.."/player.table"), true)
    local info = serializer.load(savefile.."/info.table")
---@diagnostic disable-next-line: param-type-mismatch
    for key, value in pairs(info) do
        self[key] = value
    end
    local chunkx, chunky = self:getchunkcoords(self:getspace(self.player.x, self.player.y))
    self:autoloadchunks(chunkx, chunky)
end

-- statics
world.scale = 3

function world:save(newsave)
    local savefile = newsave or self.savefile
    love.filesystem.createDirectory(savefile)
    love.filesystem.createDirectory(savefile.."/chunkdata")
    for id, chunk in pairs(self.chunks) do
        serializer.save(chunk, savefile.."/chunkdata/"..id..".table")
    end
    serializer.save({
        seed = self.seed,
        time = self.time,
        uid = self.uid,
        generator = self.generator,
        generatorsettings = self.generatorsettings
    }, savefile.."/info.table")
    serializer.save(self.objects, savefile.."/objects.table")
    serializer.save(self.metax, savefile.."/metax.table")
    serializer.save(self.metay, savefile.."/metay.table")
    serializer.save(self.player, savefile.."/player.table")
end

return function()
    return setmetatable({}, {
        __index = world
    })
end